{
    hlog - Hobo's Logging Library for Cobra
    
    ---------------------------------------
    
    Copyright (c) 2008 Ben Golightly

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation
    files (the "Software"), to deal in the Software without
    restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following
    conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
    OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
    NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
    WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
    OTHER DEALINGS IN THE SOFTWARE.

    ---------------------------------------
    
    Online Documentation:
    http://www.tophatstuff.co.uk/man/Cobra/hlog/readme.html 
    
}
Unit

Const
    HLOG_VERSION = "20080921"       ; Export
    QUIET = FALSE                               // When TRUE, the MessageBox created by Error(msg: String) is supressed. The logging system
                                                // should still run without crashing, but it might be slow if you're running into recursion errors.
    SAFE = TRUE                                 // When set to FALSE, most pointer checks are skipped. (FALSE is about 1% faster)
    RECURSION_THRESHOLD = 32                    // When this threshold is reached, it is assumed that there is an infinite loop
    ENTRY_PREPEND_TIME = FALSE                  // TRUE to prepend Log entries with a date & time string (FALSE is about 25% faster)
    ENTRY_PREPEND_MILLISECS = TRUE              // TRUE to prepend Log entries with the Millisecs since start
    INDENTATION_STRING = " "                    // String used to indent (default: a space)

Type childLoggingContexts = Record
    contextId: Integer
EndType ; Export

Type loggingContexts = Record

    { Properties }
    logFile: String
    entries: Integer // A count of how many entries have been made
    
    { Flags }
    realtime: Boolean
    enabled: Boolean
    
    { Private }
    buffer: String = ""
    indentation: Integer = 0
    
    { Lists }    
    childLoggingContextList: List of childLoggingContexts
    
EndType ; Export

Var
    loggingContextList : List of loggingContexts ; Export
    start : Integer

Function CreateLoggingContext(logFile: String, realtime: Boolean, enabled: Boolean=TRUE) :^loggingContexts ; Export
Var
    loggingContext :^loggingContexts
    err: String = ""
Begin

    loggingContext = NewItem(loggingContextList)
    If Assigned(loggingContext) Then

        If FileExists(logFile) then
            If DeleteFile(logFile) = FALSE then
                err =       "Fatal Error in CreateLoggingContext:" + Chr(10) + Chr(10)
                err = err + "The target log file already exists, but cannot be cleared." + Chr(10)
                err = err + "Do you have permission to access the file?" + Chr(10)
                err = err + "Is the file or volume read-only?" + Chr(10) + Chr(10)
                err = err + "logFile = '" + logFile + "'" + Chr(10)
                err = err + "realtime = " + realtime + Chr(10)
                err = err + "enabled = " + enabled + Chr(10) + Chr(10)
                err = err + "FileExists(logFile) = TRUE" + Chr(10)
                err = err + "DeleteFile(logFile) = FALSE"
                Error(err)
                Exit
            Endif 
        Endif
        
        loggingContext.logFile = logFile
        
        loggingContext.realtime = realtime
        loggingContext.enabled = enabled
    
        Result = loggingContext
        
    Else
        err =       "Error in CreateLoggingContext:" + Chr(10) + Chr(10)
        err = err + "A LoggingContext item could not be created. Out of memory?" + Chr(10) + Chr(10)
        err = err + "logFile = '" + logFile + "'" + Chr(10)
        err = err + "realtime = " + realtime + Chr(10)
        err = err + "enabled = " + enabled + Chr(10) + Chr(10)
        err = err + "Assigned(loggingContext) = False"                
        Error(err)
    Endif

End

Function AddChildLoggingContext(context: ^loggingContexts, add: ^loggingContexts) : Boolean ; Export
Var
    childLoggingContext : ^childLoggingContexts
    err: String = ""
Begin
    Result = TRUE

    { Check both contexts are valid }
    If SAFE then Result = CheckTwoContexts("AddChildLoggingContext", context, add)
    If Result = FALSE then Exit
    
    { Add child context Id to list }
    childLoggingContext = NewItem(context.childLoggingContextList)
    If Assigned(childLoggingContext) Then
        childLoggingContext.contextId = ItemInList(loggingContextList, add)
    Else
        err =       "Error in AddChildLoggingContext:" + Chr(10) + Chr(10)
        err = err + "A childLoggingContext item could not be created. Out of memory?" + Chr(10) + Chr(10)
        err = err + "Assigned(childLoggingContext) = False"                
        Error(err)
        Result = FALSE
        Exit
    Endif
    
    Result = TRUE

End

Function RemoveChildLoggingContext(context: ^loggingContexts, remove: ^loggingContexts) : Boolean ; Export
Var
    childLoggingContext : ^childLoggingContexts
    err: String = ""
Begin
    Result = TRUE

    { Check both contexts are valid }
    If SAFE then Result = CheckTwoContexts("RemoveChildLoggingContext", context, remove)
    If Result = FALSE then Exit
    
    { Remove child Context Id from List }
    Loop childLoggingContext through context.childLoggingContextList
        If childLoggingContext.contextId = ItemInList(loggingContextList, remove) then
            Free(childLoggingContext)
            Result = TRUE
            Exit
        Endif
    EndLoop
    
    Result = FALSE
    err =       "Error in RemoveChildLoggingContext:" + Chr(10) + Chr(10)
    err = err + "The childLoggingContext wasn't found in the parent loggingContext's list of children." + Chr(10)
    err = err + "It's likely that either the childLoggingContext was never actually a child of the parent loggingContext," + Chr(10)
    err = err + "Or that the childLoggingContext has already been removed from the parent loggingContext's child list."
    Error(err)

End

Function LogMsg(context :^loggingContexts, msg: String, indentBefore: Integer=0, indentAfter: Integer=0, overflow: Integer=0) : Boolean ; Export
Var
    childLoggingContext : ^childLoggingContexts
    child: ^loggingContexts
Begin
    Result = TRUE

    If overflow > RECURSION_THRESHOLD then
        Error_Overflow("LogTo")
        Result = FALSE
        Exit
    Endif
    
    { Check context is valid }
    If SAFE then Result = CheckContext("LogTo", context)
    If Result = FALSE then Exit
    
    If context.enabled = FALSE then exit
    
    If indentBefore <> 0 then IndentLog(context, indentBefore)

    Inc(context.entries)    
    If context.realtime = TRUE then
        LogWrite(context.logFile, FormatLogEntryString(msg, context.indentation))
    Else
        context.buffer = context.buffer + FormatLogEntryString(msg, context.indentation) + Chr(10)   
    Endif
    
    { Children }
    Loop childLoggingContext through context.childLoggingContextList
    
        child = IndexedItem(loggingContextList, childLoggingContext.contextId)
        
        If SAFE then Result = CheckContext("LogTo(child)", child)
        If Result = FALSE then Exit
        
        LogMsg(child, msg, 0, 0, overflow+1)
        
    EndLoop
        
    If indentAfter <> 0 then IndentLog(context, indentAfter)

    Result = TRUE
    
End

Function FormatLogEntryString(str: String, indentation: Integer) : String
Begin
    Result = ""
    If ENTRY_PREPEND_TIME then      Result = DateTimeToString(Now, "%yyyy:%mm%dd %hh:%nn:%ss ") + " "    
    If ENTRY_PREPEND_MILLISECS then Result = Result + RPad(Millisecs - start, 11, "")
    Result =                        Result + RepStr(INDENTATION_STRING, indentation * 4)
    Result =                        Result + str
End

Procedure LogWrite(logFile: String, msg: String)
Var
    fstream : Element
    logSize : Integer = 0
Begin
    
    If FileExists(logFile) then    
        fstream = ReadFile(logFile)
        logSize = FileSize(fstream)
        Closefile(fstream)
    Endif
    
    fstream = WriteFile(logFile)
    FileSeek(fstream,logSize)
    WriteLn(fstream, msg)
    Closefile(fstream)

End

Function IndentLog(context :^loggingContexts, indent: Integer=1, overflow: Integer=0) : Boolean ; Export
Var
    childLoggingContext : ^childLoggingContexts
    child: ^loggingContexts
    
    err : String = ""
Begin
    Result = TRUE
    If indent = 0 then Exit
    
    If overflow > RECURSION_THRESHOLD then
        Error_Overflow("IndentLog")
        Result = FALSE
        Exit
    Endif
    
    { Check context is valid }
    If SAFE then Result = CheckContext("IndentLog", context)
    If Result = FALSE then Exit
    
    If context.enabled = FALSE then exit
    
    context.indentation = context.indentation + indent
    If context.indentation < 0 then
        err =       "Error in IndentLog: Negative indentation!" + Chr(10) + Chr(10)
        err = err + "The loggingContext logFile is '"+context.logFile+"'"
        Error(err)
        Result = FALSE
        Exit
    Endif
    
    Loop childLoggingContext through context.childLoggingContextList
    
        child = IndexedItem(loggingContextList, childLoggingContext.contextId)
        
        If SAFE then Result = CheckContext("IndentLog(child)", child)
        If Result = FALSE then Exit
        
        IndentLog(child, indent, overflow+1)
        
    EndLoop
    
    Result = TRUE
     
End

Function UnindentLog(context :^loggingContexts, indent: Integer=1) : Boolean ; Export
Begin
    If indent = 0 then Exit
    Result = IndentLog(context, -indent)
End

Function CommitLogBuffer(context :^loggingContexts) : Boolean ; Export
Begin
    Result = TRUE
    
    { Check context is valid }
    If SAFE then Result = CheckContext("CommitLogBuffer",context)
    If Result = FALSE then Exit
    
    If context.enabled = FALSE then exit
    
    If context.buffer = "" then Exit
    LogWrite(context.logFile, context.buffer)
    context.buffer = ""
    
End

Function ClearLogBuffer(context :^loggingContexts) : Boolean ; Export
Begin
    Result = TRUE
    
    { Check context is valid }
    If SAFE then Result = CheckContext("ClearLogBuffer",context)
    If Result = FALSE then Exit
    
    context.buffer = ""
    
End

Function FreeLoggingContext(context: ^loggingContexts) : Boolean ; Export
Var
    childLoggingContext : ^childLoggingContexts
Begin
    Result = TRUE
    
    { Check context is valid }
    If SAFE then Result = CheckContext("FreeLoggingContext",context)
    If Result = FALSE then Exit

    If context.buffer <> "" then CommitLogBuffer(context)
    
    Loop childLoggingContext through context.childLoggingContextList
        Free(childLoggingContext)
    EndLoop

    Free(context)    
    Result = TRUE
End

Function LoggingContextRealtime(context: ^loggingContexts, realtime: Boolean) : Boolean ; Export
Begin
    Result = TRUE
    
    { Check context is valid }
    If SAFE then Result = CheckContext("LoggingContextRealtime",context)
    If Result = FALSE then Exit
    
    context.realtime = realtime

End

Function LoggingContextEnabled(context: ^loggingContexts, enabled: Boolean) : Boolean ; Export
Begin
    Result = TRUE
    
    { Check context is valid }
    If SAFE then Result = CheckContext("LoggingContextEnabled",context)
    If Result = FALSE then Exit
    
    context.enabled = enabled

End

Function FreeAllLoggingContexts() : Boolean ; Export
Var
    loggingContext :^loggingContexts
Begin
    Result = TRUE
    
    Loop loggingContext through loggingContextList    
        Result = Result AND FreeLoggingContext(loggingContext)
    EndLoop

End

Function CheckContext(funcName: String, context: ^loggingContexts) : Boolean
Var
    err : String = ""
    err2 : String = ""
Begin
    If ItemInList(loggingContextList, context) = -1 then
        err = "ItemInList(context, loggingContextList) = -1"
    Endif
    If err <> "" then
        err2 =        "Error in " + funcName + ":" + Chr(10) + Chr(10)
        err2 = err2 + "The supplied logging context was invalid." + Chr(10) + Chr(10)
        err2 = err2 + err
        Error(err2)
        Result = FALSE
        Exit
    Endif
    
    Result = TRUE
End

Function CheckTwoContexts(funcName: String, context1: ^loggingContexts, context2: ^loggingContexts) : Boolean
Var
    err : String = ""
    err2 : String = ""
Begin
    { Check both contexts are valid }
    If ItemInList(loggingContextList, context1) = -1 then
        err = "ItemInList(context1, loggingContextList) = -1"
    Endif
    If ItemInList(loggingContextList, context2) = -1 then
        If err <> "" then err = err + Chr(10)
        err = err + "ItemInList(context2, loggingContextList) = -1"
    Endif
    If err <> "" then
        err2 =        "Error in " + funcName + ":" + Chr(10) + Chr(10)
        err2 = err2 + "One or both of the supplied logging contexts were invalid." + Chr(10) + Chr(10)
        err2 = err2 + err
        Error(err2)
        Result = FALSE
        Exit
    Endif
    
    Result = TRUE
End

Procedure Error(msg: String)
Begin
    If Not QUIET then MessageBox(msg, "hlog")
End

Procedure Error_Overflow(funcName: String)
Var
    err: String = ""
Begin
    err =       "Error in "+funcName+": Infinite Recursion Detected!" + Chr(10) + Chr(10)
    err = err + "It's likely that you used this function with a loggingContext that has child" + Chr(10)
    err = err + "loggingContexts that have been freed but not removed from this" + Chr(10)
    err = err + "loggingContext's list of children with RemoveChildLoggingContext" + Chr(10) + Chr(10)
    err = err + "It's also possible that you have a loggingContext with itself as a child," + Chr(10)
    err = err + "causing an infinite loop." + Chr(10) + Chr(10)
    err = err + "Aborting!"
    Error(err)
End

Begin
    start = Millisecs
End